/*
 * FILE:
 *   xkb-greek.c
 * DESCRIPTION:
 *   Main module for the "xkb-greek" program. Xkb-greek remaps the
 *   keyboard of an XKB-capable server, for bilingual greek/latin 
 *   input.
 * AUTHOR:
 *   Nick Patavalis - (npat@di.uoa.gr)
 * MODIFIED:
 *    May 22, 1997 - initial devel.
 *    Jun 2, 1997 by npat (version 0.1 now oficially ready!)
 *       * Added the command-line parser. 
 *       * Cleaned code up a bit.
 *       * Added the IGNORE_PENALTY and IGNORE_LOCK conditionals.
 *       * Added a check to see if the alpha keys have allready 
 *       (before remapping) more than one groups assigned to. It 
 *       this case just a message is printed.
 *       * Deleted some unused varibles.
 *       * Created the version.h file.
 *    Mar 14, 1998
 *       * Added the InhibitGroups function which is *not* called
 *         yet. This is v0.2
 *       * Corrected a NASTY bug on the way toggle was handled.
 * NOTES:
 *   XKB: Who said keyboard mapping should not be rocket science?
 *
 *   For the time beign (v0.2) this program is VERY-VERY alpha. 
 *   It works on my linux box but I don't know if it will work
 *   on any other systems (it should). Please befor trying it out read 
 *   the README file. Also DO browse to the large blocks of comments 
 *   (not the little details) in the source code to get an idea what 
 *   this program is trying to do.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <X11/extensions/XKBfile.h>

#include "version.h"

/***************************************************************************/

#define DEBUG

/****** Do >not< comment >in< unless you know what you are doing!
#define IGNORE_PENALTY
#define IGNORE_LOCK
*******/

/*
 * If you have any reason to believe that your XKB implementation
 * honors entity names (as it should!) comment this out. My XKB
 * implementation (XKB v1.0 on XFree86 v3.2 does NOT). 
 * ANYWAY it is safer to live it on.
 *
 * Read the comment just before the CheckIfClean function before
 * touching it!!
 */
#define UNRELIABLE_NAMES

/*
 * Theoretically (and maybe even practically), you could increase this
 * to make the system more tolerant to penalties when selecting the
 * accent modifier. But BEWARE: by doing so you might override some
 * allready used and usefull modifier and induce some insane
 * behavior to your keyboard.
 * 
 * Read the comments just before the "FindFreeMod" function before 
 * touching this!
 */
#define PENALTY_MAX 1


/* Don't even think about touching the rest defines! */

#define LEVELS_NUMBER 6
#define MAPS_NUMBER 14
#define MAPPEDKEYS_NUM 27

#define SM (ShiftMask)
#define LM (LockMask)
#define CM (ControlMask)
#define MM (MetaMask)
#define AM (AccentMask)

/***************************************************************************/

/* Heavy pointer magic! Do not modify! It works!! */

#define locks(_x,_k,_s,_g) \
((_x)->map->syms[((_x)->map->key_sym_map[(_k)].offset) + \
				(((_x)->map->key_sym_map[(_k)].width)*(_g)) + (_s)])

#define locac(_x,_k,_s,_g) \
	((_x)->server->acts[((_x)->server->key_acts[(_k)]) + \
					   (((_x)->map->key_sym_map[(_k)].width)*(_g)) + \
					   (_s)])

/***************************************************************************/

typedef struct _mapper {
	int kk;
	int lks;
	int Lks;
	int gks;
	int Gks;
	int gaks;
	int Gaks;
} KEYBDMAPPER;

KEYBDMAPPER almap[MAPPEDKEYS_NUM];

/***************************************************************************/

char errMsg[1024] = "No Error!\n";

void 
setError (char *error) {
    strcpy(errMsg,error);
}

void 
showError (char *err) {
    if (err) strcpy(errMsg,err);
    fprintf(stderr,"** ERROR **: %s\n",errMsg);
}

/***************************************************************************/

/* This function looks ugly! So I put it in a separete file.

   Bool DoMap (Display *dpy, XkbDescPtr xkb) 
   
*/

#include "mapper.h"

/************************************************************************/


Display *
OpenXkb (int v, char *displayName, int *evt_type)
{
    Display *dpy;
    int evt_rtn, err_rtn, major_in, minor_in, why_rtn;

    major_in=1;
    minor_in=0;

    dpy = XkbOpenDisplay( displayName,
						  &evt_rtn,
						  &err_rtn,
						  &major_in,
						  &minor_in,
						  &why_rtn);
    switch (why_rtn) {
    case XkbOD_BadLibraryVersion:
		setError("Conection attempt resulted in: XkbOD_BadLibraryVersion");
		return(NULL);
    case XkbOD_ConnectionRefused:
		setError("Conection attempt resulted in: XkbOD_ConnectionRefused");
		return(NULL);
    case XkbOD_BadServerVersion:
		setError("Conection attempt resulted in: XkbOD_BadServerVersion");
		return(NULL);
    case XkbOD_NonXkbServer:
		setError("Conection attempt resulted in: XkbOD_NonXkbServer");
		return(NULL);
    case XkbOD_Success:
		break;
    default:
		assert(0);
		break;
    }

    if (v) {
		printf("XKbd server version v%d.%d. Reporting events as type #%d\n",
			   major_in, minor_in, evt_rtn);
    }

    if (evt_type) *evt_type = evt_rtn; 

    return(dpy);
}

/***************************************************************************/

XkbDescPtr xkbOrig;

/*
 * These two functions (SaveOriginalMap and RestoreOriginalMap) are
 * not at all what they should be! 
 *
 * Unfortunatelly saving and
 * restoring the keyboard state is not so easy as it may seem. Their
 * basic problem is that they are unable to restore the original
 * key-types. So if a new type is added, the type will still be
 * there even after calling Restore. I know no "safe" way of
 * restoring the "original" key-types (automagically erasing
 * the new ones). If you do please let me know. As a result
 * this program must only be run ONCE during the lifetime of
 * a Xserver-run, EVEN if the first run was not successfull.
 *
 * Saving the original keyboard mapping in a *file* is even
 * harder (although possible). It *would* be very usefull though,
 * since it *would* allow to clean and remap the keyboard, many times
 * per X-server-run.
 */
Bool 
SaveOriginalMap (Display *dpy, XkbDescPtr xkb)
{
	if (!(xkbOrig = XkbGetMap(dpy, XkbAllMapComponentsMask, XkbUseCoreKbd))) {
		return(False);
	} else {
		return(True);
	}
}

Bool 
RestoreOriginalMap(Display *dpy)
{
	if (!(XkbSetMap(dpy, XkbAllMapComponentsMask, xkbOrig))) {
		return(False);
	}
	if ( XkbGetUpdatedMap(dpy, XkbAllMapComponentsMask, xkbOrig) != Success) {
		return(False);
	}
	return(True);
}


/***************************************************************************/

/*
 * Check if this program (xkb-greek) has allready been run.
 *
 * It actually checks for the greek key-type. If present it returns
 * false. The canonical way to look for the presense of the greek
 * type is by scanning the type *names*, and looking for its name
 * (a.k.a "greek-key" or something like this). Unfortunatelly and for
 * a reason unknown to me, it seems as though the key-types (and any 
 * other xkb entity), do not "keep" their names, even if they 
 * (the names) are explicitly set. In order to circumvent this
 * problem I found no other solution but to search for key-types
 * that *resemble* the greek one! The greek type is a very 
 * perverted one (it has 6 shift-levels, and some 14 maps!) so it 
 * should be easily distinguishable among the others...  
 *
 */

Bool 
CheckIfClean (Display *dpy, XkbDescPtr xkb)
{
	int i;

	for (i=0; i<xkb->map->num_types; i++) {
		if (xkb->map->types[i].name != 0x0) {
			printf ("Reeee\n");
			if (!strcmp("greekkey",XGetAtomName(dpy,xkb->map->types[i].name)))
				return(False);
		}
#ifdef UNRELIABLE_NAMES
		if ( (xkb->map->types[i].num_levels == LEVELS_NUMBER) &&
			 (xkb->map->types[i].map_count == MAPS_NUMBER) ) 
			return(False);
#endif UNRELIABLE_NAMES	 
	}
	return(True);
}

/***************************************************************************/

/*
 * In order for the greek keyboard mapping to work, we need a clean 
 * (unused) modifier. If no such modified exists (all modifiers are
 * somehow used), a not-so-heavily-used modifier might do!!
 *
 * This "clean"-modifier is used to implement the accent behavior.
 *
 * In order to locate a suitable modifier this function scans 
 * through the complete keyboard definition and assigns penalties
 * to modifiers it finds either bound to keys or used at key-type
 * definitions.
 *
 * If a modifier is bound to a key, it takes a 1-point penalty,
 * If a modifier is used for in the definition of a key-type it gets
 *     2-points of penalty.
 *
 * If you want to AVOID all this poor-man's AI, you can specify 
 * manually the modifier (mod1 through mod5) to be used. 
 */
int FindFreeMod (Display *xpdy, XkbDescPtr xkb)
{
    int i, m, rv, mn, penalty[5];

    /* scan through type-definitions looking for modifier references.
       Assign +2 penalty to every one found. */
    for (m=Mod1Mask, mn=0; m<=Mod5Mask; m<<=1, mn++) {
		penalty[mn] = 0;
		for (i=0; i<xkb->map->num_types; i++) 
			if (xkb->map->types[i].mods.real_mods & m) penalty[mn]+=2;
    }
    
    /* scan throug key-binding add +1 penalty to any bound modifier. */
    for (m=Mod1Mask, mn=0; m<=Mod5Mask; m<<=1, mn++) {
		for (i=xkb->min_key_code; i<=xkb->max_key_code; i++)
			if (xkb->map->modmap[i] & m) {penalty[mn]+=1; printf("%d %d\n",mn+1,i);}
    }

    for (i=0, rv=0, mn=penalty[0]; i<5; i++) {
#ifdef DEBUG
		printf("penalty for mod %02d is %d:\n",i+1, penalty[i]);
#endif
		if (penalty[i] < mn) {mn = penalty[i]; rv=i; }
    }

#ifdef DEBUG
    printf("Modifer selected for accent is: %d\n",rv+1);
#endif

#ifndef IGNORE_PENALTY
    if (mn > PENALTY_MAX)
		return(-1);
    else 
#endif
		return(Mod1Mask << rv);
}

/***************************************************************************/

/*
 * This the core of the greek support. The greek keytype MONSTER!
 *
 * The greek keytype has 6 (yes! six!) shift-levels and 14 maps
 * (map = combination of modifiers that leads to a shift-level).
 *
 * Ther 6 levels are:
 *      level 0: lowcase latin char.
 *      level 1: upcase latin char.
 *      level 2: lowcase greek char.
 *      level 3: upcase greek char.
 *      level 4: lowcase accented greek char.
 *      level 5: upcase accented greek char.
 *
 * Please observe that the first two levels ARE NOT the normal latin
 * levels!!! They ARE NOT the ones used when the keyboard is in latin
 * "mode". Latin mode is a different *group*. Levels 0 and 1 are only
 * used to implement a sane CONTROL-characters behavior. So when
 * the keyboard is in greek-mode AND you type, say, CTRL-a instead
 * of CTRL-Greek_alpha the usual CTRL-a is send to the client. This
 * allows you to use control (and META) keys while in greek mode, 
 * without having to shift back to the latin mode. 
 *
 * In order to achieve this I had to do the following:
 * Set level 2 as the default greek-mode shift-level (when no modifier keys
 * are on). Then I program CTRL and META modifiers, to to shift DOWN to 
 * level 0, and when Shift (or Caps) is also pressed to shift DOWN to 
 * level 1 instead: Down there latin characters are safely mapped! I also
 * had to REGENERATE and PRESERVE the CTRL, META, and Shift modifers, 
 * that where by default consumed by this down-shifting process. Having
 * done all this the client recieves a familiar CTRL-a instead of an
 * alien-looking CTRL-Greek_alpha, even if the keyboard is in 
 * greek-mode!
 *
 * Does it work? 99% of the applications behave very well. But some 
 * very clever (or very stupid) applications, tend to test the COMPLETE 
 * keyboard state without any translation, INCLUDING the group-bits.
 * This means that for them a CTRL-a comming from a keyboard in group-1
 * is DIFFERENT than a CTRL-a comming form a keyboard in group-0. 
 * Now, this is no-questions-plain-stupid! but its how it is! 
 *
 * Bringing the group-bits down when CTRL or META are pressed WHILE 
 * preserving the modifier-bits, AND assert back the group bits when 
 * the modifier bits are released, would require some houndini-skils, 
 * if at all possible!!
 */ 

int DefineGreekType (Display *dpy, XkbDescPtr xkb, 
					 int MetaMask, int AccentMask)
{
    int i, grti;
    XkbKeyTypePtr t;

    if (xkb->map->num_types == xkb->map->size_types) {
		/* expand key-types array. */
		if ( XkbAllocClientMap(xkb, XkbKeyTypesMask, 
							   xkb->map->size_types + 1) != Success ) {
			setError("Cannot expand key-types array!");
			return(-1);
		}
    }

    /* Copy a typical base-type to the newly allocated type-slot.
       This gives us a place to start. */
    grti = xkb->map->size_types - 1;
    if ( XkbCopyKeyType(&(xkb->map->types[0]), 
						&(xkb->map->types[grti])) != Success ) {
		setError("Cannot Copy base key-type.");
		return(-1);
    }
    xkb->map->num_types++;

    /* Resize the grti'th type to the number of levels we need 
       for greek. (We need 6! levels, and 14 maps --- 
       YEP! The space kadet keyboard is back!!).*/
    if ( XkbResizeKeyType(xkb,grti,14,True,6) != Success ) {
		setError("Cannot resize greek-type.");
		return(-1);
    }

    /* set the level names. */
    xkb->map->types[grti].name = XkbInternAtom(NULL, "greekkey",0);
    xkb->map->types[grti].level_names[0] = XkbInternAtom(NULL, "grel",0);
    xkb->map->types[grti].level_names[1] = XkbInternAtom(NULL, "greu",0);
    xkb->map->types[grti].level_names[2] = XkbInternAtom(NULL, "grlow",0);
    xkb->map->types[grti].level_names[3] = XkbInternAtom(NULL, "grup",0);
    xkb->map->types[grti].level_names[4] = XkbInternAtom(NULL, "graclow",0);
    xkb->map->types[grti].level_names[5] = XkbInternAtom(NULL, "gracup",0);

    /* The following are the rules for handling the modifier bits
       in greek mode (greek key-type actually). They got a little 
       complicated down the way, but they do a good (though not 
       perfect) job. */
    
    t = &(xkb->map->types[grti]);
    t->mods.real_mods = 
		LockMask | 
		ShiftMask | 
		ControlMask | 
		MetaMask | 
		AccentMask;
    
    for (i=0; i<14; t->map[i++].active = True);
    
    t->map[0].mods.real_mods  = 0;        t->map[0].level  = 2;
    t->map[1].mods.real_mods  = SM;       t->map[1].level  = 3;
    t->map[2].mods.real_mods  = LM;       t->map[2].level  = 3;
    t->map[3].mods.real_mods  = SM|LM;    t->map[3].level  = 3;
    t->map[4].mods.real_mods  = AM;       t->map[4].level  = 4;
    t->map[5].mods.real_mods  = AM|SM;    t->map[5].level  = 5;
    t->map[6].mods.real_mods  = AM|LM;    t->map[6].level  = 5;
    t->map[7].mods.real_mods  = AM|SM|LM; t->map[7].level  = 5;
    t->map[8].mods.real_mods  = CM;       t->map[8].level  = 0;
    t->map[9].mods.real_mods  = MM;       t->map[9].level  = 0;
    t->map[10].mods.real_mods = CM|MM;    t->map[10].level = 0;
    t->map[11].mods.real_mods = CM|SM;    t->map[11].level = 1;
    t->map[12].mods.real_mods = MM|SM;    t->map[12].level = 1;
    t->map[13].mods.real_mods = CM|MM|SM; t->map[13].level = 1;
    
    
    t->preserve[0].real_mods  = 0;
    t->preserve[1].real_mods  = 0;
    t->preserve[2].real_mods  = 0;
    t->preserve[3].real_mods  = 0;
    t->preserve[4].real_mods  = 0;
    t->preserve[5].real_mods  = 0;
    t->preserve[6].real_mods  = 0;
    t->preserve[7].real_mods  = 0;
    t->preserve[8].real_mods  = CM;
    t->preserve[9].real_mods  = MM;
    t->preserve[10].real_mods = CM | MM;
    t->preserve[11].real_mods = CM | SM;
    t->preserve[12].real_mods = MM | SM;
    t->preserve[13].real_mods = CM | MM | SM;

#ifdef DEBUG
	fprintf(stderr,"Definig greek key-type...\n");
	if ( (!XkbSetMap(dpy, XkbAllMapComponentsMask, xkb))
		 || (XkbGetUpdatedMap(dpy, XkbAllMapComponentsMask, xkb)!=Success) 
		) {
		setError("Cannot modify server-tables.");
		return(-1);
    }
#endif

    return(grti);
}

/***************************************************************************/

Bool RemapKeyboard (Display *dpy, XkbDescPtr xkb, int grti, 
					int ToggleKey)
{
    int i, grouptypes[4];

    /* Assign two groups to the alphanumerical keys and
       remap the keysyms acordingly. */
    for(i=0; i<26; i++) {
#ifdef DEBUG
		if (XkbKeyNumGroups(xkb, almap[i].kk) != 1) {
			printf("WARNING: Overriding group(s) for kcode: %d\n",
				   almap[i].kk);
		}
#endif
		/* keep the original key-type for group 0. */
		grouptypes[0]=xkb->map->key_sym_map[almap[i].kk].kt_index[0];
		/* set the key-type to greek for group 1. */
		grouptypes[1]=grti;
		if ( XkbChangeTypesOfKey(xkb, almap[i].kk, 
								 2, XkbGroup1Mask | XkbGroup2Mask,
								 grouptypes,
								 NULL)!=Success) {
			setError("Cannot change type of alpha key!\n");
			return(False);
		}
		/* now assing the coresponding key-syms to all 6 levels of 
		   the greek type. */
		locks(xkb,(almap[i].kk),0,1) = almap[i].lks;
		locks(xkb,(almap[i].kk),1,1) = almap[i].Lks;
		locks(xkb,(almap[i].kk),2,1) = almap[i].gks;
		locks(xkb,(almap[i].kk),3,1) = almap[i].Gks;
		locks(xkb,(almap[i].kk),4,1) = almap[i].gaks;
		locks(xkb,(almap[i].kk),5,1) = almap[i].Gaks;
    }

#ifdef DEBUG
	fprintf(stderr,"Remaping alphabetic keys...\n");
	if ( (!XkbSetMap(dpy, XkbAllMapComponentsMask, xkb))
		 || (XkbGetUpdatedMap(dpy, XkbAllMapComponentsMask, xkb)!=Success) 
		) {
		setError("Cannot modify server-tables.");
		return(False);
    }
#endif
    
    /* This is for the semicolon key. */
    grouptypes[0]=xkb->map->key_sym_map[almap[SEMICOL_NDX].kk].kt_index[0];
    grouptypes[1]=XkbTwoLevelIndex;
    if ( XkbChangeTypesOfKey(xkb, almap[SEMICOL_NDX].kk, 
							 2, XkbGroup1Mask | XkbGroup2Mask,
							 grouptypes,
							 NULL)!=Success) {
		setError("Cannnot change type of semicolon key!");
		return(False);
	}
    locks(xkb,(almap[26].kk),0,1) = 0;
    locks(xkb,(almap[26].kk),1,1) = almap[26].Lks;

#ifdef DEBUG
	fprintf(stderr,"Remaping semicolon key...\n");
	if ( (!XkbSetMap(dpy, XkbAllMapComponentsMask, xkb))
		 || (XkbGetUpdatedMap(dpy, XkbAllMapComponentsMask, xkb)!=Success) 
		) {
		setError("Cannot modify server-tables.");
		return(False);
    }
#endif

    /* This is for the toggle key. */
    grouptypes[XkbGroup1Index]=XkbOneLevelIndex;
	XkbResizeKeyActions(xkb,ToggleKey,0);
	XkbResizeKeySyms(xkb,ToggleKey,0);
    if ( XkbChangeTypesOfKey(xkb, ToggleKey,
							 1, XkbGroup1Mask,
							 grouptypes,
							 NULL)!=Success) {
		return(False);
    }
    locks(xkb,ToggleKey,0,0) = XK_VoidSymbol;	

#ifdef DEBUG
	fprintf(stderr,"Remaping toggle key...\n");
	if ( (!XkbSetMap(dpy, XkbAllMapComponentsMask, xkb))
		 || (XkbGetUpdatedMap(dpy, XkbAllMapComponentsMask, xkb)!=Success) 
		) {
		setError("Cannot modify server-tables.");
		return(False);
    }
#endif

    return(True);
}

/***************************************************************************/


/*
 * NOTE: Actions are assigned to KEYSYMS, keysyms are assigned to
 * keycodes. It is NOT the same thing as if actions where assigned
 * to keycodes! (Oh! God help us!)
 */
Bool DefineActions (Display *dpy, XkbDescPtr xkb, int grti, 
					int ToggleKey, int AccentMask)
{
    /* Remove all compatibily modes from the keys to be used
       as Group togglers and Accent Modifiers. They just get 
       in our way. Also make sure they don't have any stange
       "behavior" assigned to them. */
    xkb->server->explicit[ToggleKey] = 255;
    xkb->server->explicit[almap[SEMICOL_NDX].kk] = 255;
    xkb->server->behaviors[almap[26].kk].type = 0;
    xkb->server->behaviors[ToggleKey].type = 0;

    /* Make sure that the toggle-key has two action-slots. (One
       for each level)
       If not --> resize. */
    if ( (!XkbKeyHasActions(xkb,ToggleKey)) || 
		 (XkbKeyNumActions(xkb,ToggleKey) < 1)) {
#ifdef DEBUG
		fprintf(stderr,"resizing actions for toggle key.\n");
#endif
		XkbResizeKeyActions(xkb,ToggleKey,1);
    }

    /* The action for toggle-key must be LockGroups: Plain and simple
       with no strange action-modifier flags set (does anybody 
       understainds them anyhow? --- including their designed) */
    (locac(xkb, (ToggleKey), 0, 0)).group.type = XkbSA_LockGroup;
    (locac(xkb, (ToggleKey), 0, 0)).group.flags = 0;
    (locac(xkb, (ToggleKey), 0, 0)).group.group_XXX = 1;
    
    /* This is more like an xmodmap-ism. It may not be 
       nesecairy, but with XKB, you'd better be safe than sorry! */
    xkb->map->modmap[almap[SEMICOL_NDX].kk] = AccentMask;
    
    /* Make sure that the accent-key (semicolon) has four action-slots.
       (Two for each group --- one for each level).
       If not --> resize. */
    if ( (!XkbKeyHasActions(xkb,almap[SEMICOL_NDX].kk)) || 
		 (XkbKeyNumActions(xkb,almap[SEMICOL_NDX].kk) != 4) ) {
#ifdef DEBUG
		fprintf(stderr,"resizing actions for semicolon key.\n");
#endif
		XkbResizeKeyActions(xkb,almap[SEMICOL_NDX].kk,4);
    }
    
    (locac(xkb, (almap[26].kk), 0, 1)).mods.type = XkbSA_LatchMods;
    (locac(xkb, (almap[26].kk), 0, 1)).mods.flags = 0;
    (locac(xkb, (almap[26].kk), 0, 1)).mods.mask = AccentMask;
    (locac(xkb, (almap[26].kk), 0, 1)).mods.real_mods = AccentMask;

#ifdef DEBUG
	fprintf(stderr,"Setting actions for toggle and semicolon keys...\n");
	if ( (!XkbSetMap(dpy, XkbAllMapComponentsMask, xkb))
		 || (XkbGetUpdatedMap(dpy, XkbAllMapComponentsMask, xkb)!=Success) 
		) {
		setError("Cannot modify server-tables.");
		return(False);
    }
#endif

    return(True);
}

/**************************************************************************/

Bool InhibitGroups (Display *dpy, XkbDescPtr xkb)
{
    if (XkbGetCompatMap(dpy,XkbAllCompatMask,xkb)!=Success) {
		setError("Cannot get compat map!");
		return(False);
	}
    
    xkb->compat->groups[0].mask = 0;
    xkb->compat->groups[0].real_mods = 0;
    xkb->compat->groups[0].vmods = 0;
    
    xkb->compat->groups[1].mask = 0;
    xkb->compat->groups[1].real_mods = 0;
    xkb->compat->groups[1].vmods = 0;
    
    if (XkbSetCompatMap(dpy,XkbAllCompatMask,xkb,1)!=Success) {
		setError("Cannot set compat map!");
		return(False);
	}
    
    return(True);
}

/**************************************************************************/

void PrintVersion (char *s)
{
    printf("Program:      %s v%d.%d\n", 
		   VERSION_PROGRAM, VERSION_MAJOR, VERSION_MINOR);
    printf("Developed by: %s\n", VERSION_AUTHOR);
    printf("Licence:      %s\n", VERSION_LICENCE);
    printf("Modified:     %s\n",VERSION_MODDATE);
}

/**************************************************************************/

void PrintUsage (char *s)
{
    char *s1 = NULL;

    s1 = strrchr(s, '/');
    if (!s1) s1 = s;
    
    printf("Usage is: \n" 
		   "    %s -toggle <toggle-key-code> -metamod <meta-modifier>\n"
		   "              [-accentmod <accent-modifier>]\n"
		   "              [-help] [-version]\n", s1);
    printf("Where:\n");
    printf("<togle-key-code>  is the keycode for the key you wish to use for\n"
		   "                  toggling Greek and Latin modes\n");
    printf("<meta-modifier>   the modifier bound to you META key (ALT key).\n"
		   "                  Usually is modifier 1, but could be "
		   "any of (1..5).\n");
    printf("<accent-modifier> the modifier to be used for accenting.\n"
		   "                  this can be any *unused* modifier (1..5)\n"
		   "                  If not specified the program will try to guess\n"
		   "                  it for you.\n");
    printf("-help             Show this screen.\n");
    printf("-version          Show program version and copyright info.\n");
}


/***************************************************************************/

struct CMDARGS {
    int ToggleKey;
    int MetaMod;
    int AccentMask;
};
    
void ParseCmdLine (int argc, char *argv[], struct CMDARGS *cmd)
{
    int i = 0;

    cmd->ToggleKey = cmd->MetaMod = cmd->AccentMask = 0;

    for (i=1; i<argc; i++) {
		if ( (!strcmp(argv[i],"-toggle")) && i < (argc-1)) {
			cmd->ToggleKey = atoi(argv[++i]);
		} else if ((!strcmp(argv[i],"-metamod")) && i < (argc-1)) {
			cmd->MetaMod = Mod1Mask << (atoi(argv[++i]) - 1);
		} else if ((!strcmp(argv[i],"-freemod")) && i < (argc-1)) {
			cmd->AccentMask = Mod1Mask << (atoi(argv[++i]) - 1);
		} else if (!strcmp(argv[i],"-help")) {
			PrintUsage(argv[0]);
			exit(0);
		} else if (!strcmp(argv[i],"-version")) {
			PrintVersion(argv[0]);
			exit(0);
		} else {
			PrintUsage(argv[0]);
			exit(1);
		}
    }

    if ((cmd->ToggleKey == 0) || (cmd->MetaMod == 0)) {
		PrintUsage(argv[0]);
		exit(1);
    }
}

/***************************************************************************/

int main (int argc, char *argv[])
{
    Display *dpy;
    
    XkbDescPtr xkb;
    
    struct CMDARGS cmd;
    int grti;

    ParseCmdLine(argc, argv, &cmd);

    if ( !(dpy = OpenXkb(1, "localhost:0.0", NULL)) ) {
		showError(NULL);
		exit(1);
    }

    if (!SaveOriginalMap(dpy, xkb)) {
		showError("The forces of Evil are strong today! Quiting!");
		exit(1);
    }

    xkb = XkbGetMap(dpy, XkbAllMapComponentsMask, XkbUseCoreKbd);

    if (!CheckIfClean(dpy,xkb)) {
		showError("This program has allready been run on this system! " 
				  "Please reset the keyboard and try again...\n");
#ifndef IGNORE_LOCK
		exit(1);
#endif
    }

    if (xkb->map->num_types >= XkbMaxKeyTypes-2) {
		showError("Whoa! 255 key TYPES(!!!) You do have a complex keyboard! " 
				  "No can do!");
		exit(1);
    }

    if (!cmd.AccentMask) {
		if ( (cmd.AccentMask = FindFreeMod(dpy, xkb)) < 0 ) {
			showError("Cannot locate a FREE keyboard modifier! " 
					  "Please define it manually.");
			exit(1);
		}
    }

    if (!DoMap(dpy, xkb)) {
		showError("You keyboard is missing some alphanumeric keys! "
				  "Put them back!");
		exit(1);
    }

    if ((grti = DefineGreekType(dpy, xkb, cmd.MetaMod, 
								cmd.AccentMask)) < 0)  {
		showError(NULL);
		exit(1);
    }

    if ( !RemapKeyboard(dpy,xkb,grti,cmd.ToggleKey) ) {
		showError(NULL);
		exit(1);
    }

    if ( !DefineActions(dpy,xkb,grti,cmd.ToggleKey,cmd.AccentMask) ) {
		showError(NULL);
		exit(1);
    }

#if 0
    if ( !InhibitGroups(dpy,xkb) ) {
		showError(NULL);
		exit(1);
    }
#endif

    if (!(XkbSetMap(dpy, XkbAllMapComponentsMask, xkb))) {
		showError("Cannot modify server-tables.");
		exit(1);
    }
    XkbGetUpdatedMap(dpy, XkbAllMapComponentsMask, xkb);

    XFlush(dpy);
    XFlush(dpy);

    exit(0);
}

